
if (typeof(IS_DUKTAPE) !== undefined && typeof(console) === 'undefined') console = { log: function() { print(Array.prototype.join.call(arguments, ' ')); }, warn: function() { print(Array.prototype.join.call(arguments, ' ')); } };
/*
Copyright 2014 David Bau.

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

(function (pool, math) {
//
// The following constants are related to IEEE 754 limits.
//
var global = this,
    width = 256,        // each RC4 output is 0 <= x < 256
    chunks = 6,         // at least six RC4 outputs for each double
    digits = 52,        // there are 52 significant digits in a double
    rngname = 'random', // rngname: name for Math.random and Math.seedrandom
    startdenom = math.pow(width, chunks),
    significance = math.pow(2, digits),
    overflow = significance * 2,
    mask = width - 1,
    nodecrypto;         // node.js crypto module, initialized at the bottom.

//
// seedrandom()
// This is the seedrandom function described above.
//
function seedrandom(seed, options, callback) {
  var key = [];
  options = (options == true) ? { entropy: true } : (options || {});

  // Flatten the seed string or build one from local entropy if needed.
  var shortseed = mixkey(flatten(
    options.entropy ? [seed, tostring(pool)] :
    (seed == null) ? autoseed() : seed, 3), key);

  // Use the seed to initialize an ARC4 generator.
  var arc4 = new ARC4(key);

  // This function returns a random double in [0, 1) that contains
  // randomness in every bit of the mantissa of the IEEE 754 value.
  var prng = function() {
    var n = arc4.g(chunks),             // Start with a numerator n < 2 ^ 48
        d = startdenom,                 //   and denominator d = 2 ^ 48.
        x = 0;                          //   and no 'extra last byte'.
    while (n < significance) {          // Fill up all significant digits by
      n = (n + x) * width;              //   shifting numerator and
      d *= width;                       //   denominator and generating a
      x = arc4.g(1);                    //   new least-significant-byte.
    }
    while (n >= overflow) {             // To avoid rounding up, before adding
      n /= 2;                           //   last byte, shift everything
      d /= 2;                           //   right using integer math until
      x >>>= 1;                         //   we have exactly the desired bits.
    }
    return (n + x) / d;                 // Form the number within [0, 1).
  };

  prng.int32 = function() { return arc4.g(4) | 0; }
  prng.quick = function() { return arc4.g(4) / 0x100000000; }
  prng.double = prng;

  // Mix the randomness into accumulated entropy.
  mixkey(tostring(arc4.S), pool);

  // Calling convention: what to return as a function of prng, seed, is_math.
  return (options.pass || callback ||
      function(prng, seed, is_math_call, state) {
        if (state) {
          // Load the arc4 state from the given state if it has an S array.
          if (state.S) { copy(state, arc4); }
          // Only provide the .state method if requested via options.state.
          prng.state = function() { return copy(arc4, {}); }
        }

        // If called as a method of Math (Math.seedrandom()), mutate
        // Math.random because that is how seedrandom.js has worked since v1.0.
        if (is_math_call) { math[rngname] = prng; return seed; }

        // Otherwise, it is a newer calling convention, so return the
        // prng directly.
        else return prng;
      })(
  prng,
  shortseed,
  'global' in options ? options.global : (this == math),
  options.state);
}
math['seed' + rngname] = seedrandom;

//
// ARC4
//
// An ARC4 implementation.  The constructor takes a key in the form of
// an array of at most (width) integers that should be 0 <= x < (width).
//
// The g(count) method returns a pseudorandom integer that concatenates
// the next (count) outputs from ARC4.  Its return value is a number x
// that is in the range 0 <= x < (width ^ count).
//
function ARC4(key) {
  var t, keylen = key.length,
      me = this, i = 0, j = me.i = me.j = 0, s = me.S = [];

  // The empty key [] is treated as [0].
  if (!keylen) { key = [keylen++]; }

  // Set up S using the standard key scheduling algorithm.
  while (i < width) {
    s[i] = i++;
  }
  for (i = 0; i < width; i++) {
    s[i] = s[j = mask & (j + key[i % keylen] + (t = s[i]))];
    s[j] = t;
  }

  // The "g" method returns the next (count) outputs as one number.
  (me.g = function(count) {
    // Using instance members instead of closure state nearly doubles speed.
    var t, r = 0,
        i = me.i, j = me.j, s = me.S;
    while (count--) {
      t = s[i = mask & (i + 1)];
      r = r * width + s[mask & ((s[i] = s[j = mask & (j + t)]) + (s[j] = t))];
    }
    me.i = i; me.j = j;
    return r;
    // For robust unpredictability, the function call below automatically
    // discards an initial batch of values.  This is called RC4-drop[256].
    // See http://google.com/search?q=rsa+fluhrer+response&btnI
  })(width);
}

//
// copy()
// Copies internal state of ARC4 to or from a plain object.
//
function copy(f, t) {
  t.i = f.i;
  t.j = f.j;
  t.S = f.S.slice();
  return t;
};

//
// flatten()
// Converts an object tree to nested arrays of strings.
//
function flatten(obj, depth) {
  var result = [], typ = (typeof obj), prop;
  if (depth && typ == 'object') {
    for (prop in obj) {
      try { result.push(flatten(obj[prop], depth - 1)); } catch (e) {}
    }
  }
  return (result.length ? result : typ == 'string' ? obj : obj + '\0');
}

//
// mixkey()
// Mixes a string seed into a key that is an array of integers, and
// returns a shortened string seed that is equivalent to the result key.
//
function mixkey(seed, key) {
  var stringseed = seed + '', smear, j = 0;
  while (j < stringseed.length) {
    key[mask & j] =
      mask & ((smear ^= key[mask & j] * 19) + stringseed.charCodeAt(j++));
  }
  return tostring(key);
}

//
// autoseed()
// Returns an object for autoseeding, using window.crypto and Node crypto
// module if available.
//
function autoseed() {
  try {
    if (nodecrypto) { return tostring(nodecrypto.randomBytes(width)); }
    var out = new Uint8Array(width);
    (global.crypto || global.msCrypto).getRandomValues(out);
    return tostring(out);
  } catch (e) {
    var browser = global.navigator,
        plugins = browser && browser.plugins;
    return [+new Date, global, plugins, global.screen, tostring(pool)];
  }
}

//
// tostring()
// Converts an array of charcodes to a string
//
function tostring(a) {
  return String.fromCharCode.apply(0, a);
}

//
// When seedrandom.js is loaded, we immediately mix a few bits
// from the built-in RNG into the entropy pool.  Because we do
// not want to interfere with deterministic PRNG state later,
// seedrandom will not call math.random on its own again after
// initialization.
//
mixkey(math.random(), pool);

//
// Nodejs and AMD support: export the implementation as a module using
// either convention.
//
if ((typeof module) == 'object' && module.exports) {
  module.exports = seedrandom;
  // When in node.js, try using crypto package for autoseeding.
  try {
    nodecrypto = require('crypto');
  } catch (ex) {}
} else if ((typeof define) == 'function' && define.amd) {
  define(function() { return seedrandom; });
}

// End anonymous scope, and pass initial values.
})(
  [],     // pool: entropy pool starts empty
  Math    // math: package containing random, pow, and seedrandom
);
var Action = (function () {
    function Action() {
    }
    Action.prototype.deserialize = function (input) {
        this.full = input.full;
        this.foreground = input.foreground;
        this.short = input.short;
        this.destructive = input.destructive;
        this.title = input.title;
        this.identifier = input.identifier;
        return this;
    };
    Action.prototype.serialize = function () {
        var output = {};
        if (this.full != null)
            output.full = this.full;
        if (this.foreground != null)
            output.foreground = this.foreground;
        if (this.short != null)
            output.short = this.short;
        if (this.destructive != null)
            output.destructive = this.destructive;
        output.title = this.title;
        output.identifier = this.identifier;
        return output;
    };
    Action.prototype.isDestructive = function () {
        return this.destructive != null && this.destructive;
    };
    return Action;
})();
var Category = (function () {
    function Category() {
    }
    Category.prototype.deserialize = function (input) {
        this.identifier = input.identifier;
        this.actions = new Array();
        if (input.actions) {
            for (var i = 0; i < input.actions.length; ++i) {
                var action = new Action();
                action = action.deserialize(input.actions[i]);
                if (action != null) {
                    this.actions.push(action);
                }
            }
        }
        return this;
    };
    Category.prototype.serialize = function () {
        var output = {};
        output.identifier = this.identifier;
        if (this.actions) {
            output.actions = [];
            for (var i = 0; i < this.actions.length; ++i) {
                output.actions.push(this.actions[i].serialize());
            }
        }
        return output;
    };
    Category.transformCategories = function (dataObject) {
        var categories = {};
        for (var i = 0; i < dataObject.length; ++i) {
            var category = new Category().deserialize(dataObject[i]);
            if (category) {
                categories[category.identifier] = category;
            }
        }
        return categories;
    };
    return Category;
})();
;
var Engine = (function () {
    function Engine(categories, waypoints, strings) {
        this._categories = categories;
        this._strings = strings;
        this._waypoints = waypoints == null ? {} : waypoints;
        this.updateRandom();
        this._gameData = new GameData();
        this._playerData = new PlayerData();
    }
    ;
    // ************************ Loading ************************
    Engine.prototype.setLanguage = function (language) {
        this._playerData.setLanguage(language);
    };
    Engine.prototype.setGameOver = function (newState) {
        return this._gameData.gameOverState(newState);
    };
    Engine.prototype.getGameOver = function () {
        return this._gameData.gameOverState();
    };
    // Update the categories and clear the waypoints before changing language
    Engine.prototype.changeLanguage = function (language, waypoints) {
        // optimizing because this is gonna take some time.
        if (this._playerData.settings.language == language) {
            return false;
        }
        var rewindContext = new RewindContext();
        rewindContext.replay = true;
        // clear out old waypoints.
        this._waypoints = {};
        if (waypoints != null) {
            this.loadWaypoints(waypoints);
        }
        var lastMadeChoice = this._playerData._simpleSequence[this._playerData._simpleSequence.length - 1];
        this._playerData.setLanguage(language);
        return this.rewindToWaypoint(lastMadeChoice.identifier, rewindContext);
    };
    Engine.prototype.loadCategories = function (dataObject) {
        elleLogger.info("Engine.loadCategories start");
        var categories = Category.transformCategories(dataObject);
        if (!categories) {
            elleLogger.warn("Failed to load categories.");
            return false;
        }
        this._categories = categories;
        elleLogger.info("Loaded categories: ", this._categories);
        return true;
    };
    Engine.prototype.loadWaypoints = function (dataObject) {
        elleLogger.info("Engine.loadWaypoints start");
        var waypointIdentifier;
        var waypointMap = Engine.waypointMapFromDataObject(dataObject);
        elleLogger.info("Engine.loadWaypoints waypoints loaded");
        if (waypointMap == null || Object.keys(waypointMap).length == 0) {
            elleLogger.warn("Failed to load waypoints from object:", dataObject);
            return false;
        }
        else {
            for (waypointIdentifier in waypointMap) {
                elleLogger.info("Loaded waypoint: ", waypointIdentifier);
                this._waypoints[waypointIdentifier] = waypointMap[waypointIdentifier];
            }
            elleLogger.info("Loaded waypoints: ", waypointMap);
        }
        if (this._waypointWaitingContext != null) {
            waypointIdentifier = this._waypointWaitingContext.waypointIdentifier;
            if (this._waypoints[waypointIdentifier] == null) {
                elleLogger.warn("We were waiting on waypoint '" + waypointIdentifier + "' but we didn't receive it");
            }
            else {
                var context = this._waypointWaitingContext.context;
                this._waypointWaitingContext = null;
                elleLogger.info("We were waiting for " + waypointIdentifier + ", let's use it.");
                this.triggerWaypointInternal(waypointIdentifier, context);
            }
        }
        elleLogger.info("Engine.loadWaypoints end");
        return true;
    };
    Engine.prototype.loadStrings = function (dataObject) {
        this._strings = dataObject;
        elleLogger.info("Loaded strings: ", this._strings);
        return true;
    };
    Engine.waypointMapFromDataObject = function (dataObject) {
        var waypoints = {};
        var waypointIdentifiers = Object.keys(dataObject);
        for (var i = 0; i < waypointIdentifiers.length; ++i) {
            var rawNodes = dataObject[waypointIdentifiers[i]];
            if (rawNodes == null) {
                elleLogger.warn("Could not find node " + waypointIdentifiers[i] + " in data object", dataObject);
                continue;
            }
            // elleLogger.info("Found " + rawNodes.length + " nodes in waypoint " + waypointIdentifiers[i]);
            var nodes = [];
            for (var j = 0; j < rawNodes.length; ++j) {
                var sequenceNode = SequenceNode.nodeFromSerializedObject(rawNodes[j]);
                if (sequenceNode) {
                    nodes.push(sequenceNode);
                }
                else {
                    elleLogger.warn("Failed to deserialize: ", rawNodes[j]);
                }
            }
            waypoints[waypointIdentifiers[i]] = nodes;
        }
        return waypoints;
    };
    Engine.prototype.getGameData = function () {
        return this._gameData.serialize();
    };
    Engine.prototype.setGameData = function (dataObject) {
        this._gameData = new GameData().deserialize(dataObject);
        return (this._gameData != null);
    };
    Engine.prototype.getPlayerData = function () {
        return this._playerData.serialize();
    };
    Engine.prototype.setPlayerData = function (dataObject) {
        this._playerData = new PlayerData().deserialize(dataObject);
        return (this._playerData != null);
    };
    Engine.prototype.getCategoryMap = function () {
        return this._categories;
    };
    Engine.prototype.getGameSpeed = function () {
        return Settings.getGameSpeedName(this._gameData.gameSpeed);
    };
    Engine.prototype.setGameSpeed = function (valueName) {
        var requestedSpeed = GameSpeed[valueName];
        if (requestedSpeed == null) {
            elleLogger.warn("Unknown requested speed: '" + valueName + "'");
            return false;
        }
        this._gameData.gameSpeed = requestedSpeed;
        return true;
    };
    // ************************ Random ************************
    Engine.prototype.updateRandom = function (seed) {
        if (!seed)
            seed = "" + Math.random();
        this._activeSeed = seed;
        this._math = new Math.seedrandom(seed);
    };
    ;
    Engine.prototype.random = function (a, b) {
        var min, max, result;
        if (a != null && b != null) {
            min = a;
            max = b;
        }
        else if (a != null) {
            max = a;
        }
        if (min != null && max != null) {
            var range = max - min;
            result = (Math.abs(this._math.int32()) % (range + 1)) + min; // inclusive range!
        }
        else if (max != null) {
            result = Math.abs(this._math.int32()) % (max + 1); // inclusive range!
        }
        else {
            result = Math.abs(this._math.int32());
        }
        return result;
    };
    Engine.prototype.either = function () {
        var values = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            values[_i - 0] = arguments[_i];
        }
        if (values.length === 0) {
            return undefined;
        }
        else if (values.length === 1) {
            return values[0];
        }
        else {
            return values[this.random(0, values.length - 1)];
        }
    };
    Engine.prototype.visited = function (waypointName, context) {
        if (context == null)
            context = new TriggerWaypointContext();
        var playerData = (context.playerData == null) ? this._playerData : context.playerData;
        if (!playerData) {
            elleLogger.warn("Missing player data");
            return -1;
        }
        var sequence = playerData._sequence;
        var count = waypointName == "Start" ? 1 : 0;
        for (var i = sequence.length - 1; i >= 0; --i) {
            var sequenceNode = sequence[i];
            if ((sequenceNode.type == NodeType.waypoint && sequenceNode.processedValue == waypointName) ||
                (sequenceNode.type == NodeType.choice && sequenceNode.input_value == waypointName)) {
                count += 1;
            }
        }
        return count;
    };
    Engine.prototype.getMusic = function () {
        return this._gameData.music;
    };
    Engine.prototype.getStringForKey = function (key) {
        return this._strings[key];
    };
    ;
    // ************************ Waypoints ************************
    Engine.prototype.rewindToBeginning = function (context) {
        if (context == null)
            context = new RewindContext();
        context = context.clone();
        if (!context.playerData)
            context.playerData = this._playerData;
        if (!context.playerData) {
            elleLogger.warn("Missing player data");
            return false;
        }
        if (!context.gameData)
            context.gameData = this._gameData;
        if (!context.gameData) {
            elleLogger.warn("Missing game data");
            return false;
        }
        context.gameData.story = {};
        context.gameData.music = undefined;
        context.playerData._sequence = [];
        context.playerData._simpleSequence = [];
        return true;
    };
    Engine.prototype.rewindToChapter = function (chapterName, context) {
        if (context == null)
            context = new RewindContext();
        context = context.clone();
        if (!context.playerData)
            context.playerData = this._playerData;
        if (!context.playerData) {
            elleLogger.warn("Missing player data");
            return false;
        }
        var simpleSequence = context.playerData._simpleSequence;
        var destinationWaypointIndentifier;
        for (var i = 0; i < simpleSequence.length; ++i) {
            if (simpleSequence[i].chapterName === chapterName) {
                destinationWaypointIndentifier = simpleSequence[i].identifier;
                break;
            }
        }
        if (destinationWaypointIndentifier == null) {
            elleLogger.warn("Chapter '" + chapterName + "' not found in simple sequence");
            return false;
        }
        var success = this.rewindToWaypoint(destinationWaypointIndentifier, context);
        if (!success)
            elleLogger.info("Found chapter but rewind to '" + destinationWaypointIndentifier + "' failed");
        return success;
    };
    Engine.prototype.rewindToChoice = function (choice, context) {
        if (context == null)
            context = new RewindContext();
        context = context.clone();
        if (!context.playerData)
            context.playerData = this._playerData;
        if (!context.playerData) {
            elleLogger.warn("Missing player data");
            return false;
        }
        var simpleSequence = context.playerData._simpleSequence;
        var destinationWaypointIndentifier;
        for (var i = 0; i < simpleSequence.length; ++i) {
            if (simpleSequence[i].identifier === choice) {
                destinationWaypointIndentifier = simpleSequence[i - 1].identifier;
                break;
            }
        }
        if (destinationWaypointIndentifier == null) {
            elleLogger.warn("Could not locate waypoint for choice '" + choice + "' in simple sequence");
            return false;
        }
        else {
            this.rewindToWaypoint(destinationWaypointIndentifier, context);
        }
        return true;
    };
    Engine.prototype.collapseLastDelay = function () {
        // get last simple sequence node:
        var simpleSequence = this._playerData._simpleSequence;
        var lastSimpleSequenceNode = simpleSequence[simpleSequence.length - 1];
        var cntNodes = lastSimpleSequenceNode.sequenceLength; // length of sequence after this simple seq added.
        if (simpleSequence.length > 1) {
            var prev = simpleSequence[simpleSequence.length - 2];
            cntNodes -= prev.sequenceLength;
        }
        var sequence = this._playerData._sequence;
        var delayNodeFound = false;
        var timeWithOffset = 0;
        for (var idx = sequence.length - cntNodes; idx < sequence.length; idx++) {
            var node = sequence[idx];
            if (timeWithOffset == 0) {
                timeWithOffset = node.availableTime;
            }
            var delay = this.delayForNode(node, this._gameData.gameSpeed, this._playerData.settings.language);
            if (!delayNodeFound && (node.type == NodeType.delay || node.type == NodeType.addelay)) {
                timeWithOffset = new Date().getTime();
                delay = 0;
                delayNodeFound = true;
            }
            if (delayNodeFound) {
                node.availableTime = timeWithOffset + delay;
                timeWithOffset += delay;
            }
        }
        return true;
    };
    Engine.prototype.rewindToWaypoint = function (waypointIdentifier, context) {
        if (context == null)
            context = new RewindContext();
        context = context.clone();
        if (!context.waypoints)
            context.waypoints = this._waypoints;
        if (!context.waypoints) {
            elleLogger.warn("Missing waypoints");
            return false;
        }
        if (!context.playerData)
            context.playerData = this._playerData;
        if (!context.playerData) {
            elleLogger.warn("Missing player data");
            return false;
        }
        if (!context.gameData)
            context.gameData = this._gameData;
        if (!context.gameData) {
            elleLogger.warn("Missing game data");
            return false;
        }
        if (!context.categories)
            context.categories = this._categories;
        if (!context.categories) {
            elleLogger.warn("Missing categories");
            return false;
        }
        if (context.replay == true) {
            var found = false;
            context.gameData.initializeStory();
            context.gameData.music = undefined;
            var originalSimpleSequence = context.playerData._simpleSequence;
            context.playerData._sequence = [];
            context.playerData._simpleSequence = [];
            for (var i = 0; i < originalSimpleSequence.length; ++i) {
                var simpleWaypoint = originalSimpleSequence[i];
                var triggerContext = new TriggerWaypointContext();
                triggerContext.seed = simpleWaypoint.seed;
                triggerContext.time = simpleWaypoint.activationEpoch;
                triggerContext.choice = simpleWaypoint.choice;
                triggerContext.waypoints = context.waypoints;
                triggerContext.categories = context.categories;
                triggerContext.playerData = context.playerData;
                triggerContext.gameData = context.gameData;
                this.triggerWaypointInternal(simpleWaypoint.identifier, triggerContext);
                if (simpleWaypoint.identifier === waypointIdentifier) {
                    found = true;
                    break;
                }
            }
            if (!found) {
                elleLogger.warn("Never found waypoint '" + waypointIdentifier + "' in the simple sequence.");
            }
        }
        else {
            elleLogger.info("Direct rewinding to " + waypointIdentifier);
            var index = 0;
            var targetSimpleWaypoint;
            for (; index < context.playerData._simpleSequence.length; ++index) {
                targetSimpleWaypoint = context.playerData._simpleSequence[index];
                if (targetSimpleWaypoint.identifier === waypointIdentifier) {
                    break;
                }
            }
            if (index === (context.playerData._simpleSequence.length - 1)) {
                elleLogger.info("Rewinding to current waypoint. Nothing to do.");
                return true;
            }
            context.playerData._simpleSequence = context.playerData._simpleSequence.slice(0, index + 1);
            context.playerData._sequence = context.playerData._sequence.slice(0, targetSimpleWaypoint.sequenceLength);
            var restoredGameData = new GameData().deserialize(targetSimpleWaypoint.serializedGameData);
            context.gameData.story = JSON.parse(JSON.stringify(restoredGameData.story));
            context.gameData.music = restoredGameData.music;
            // Clear the value of the final node
            var finalNode = context.playerData.getLastChoice();
            finalNode.input_value = undefined;
            if (finalNode.type === NodeType.text_input) {
                finalNode.evaluateNodeWithReturnType(null, this, context.gameData, true);
            }
        }
        return true;
    };
    Engine.prototype.triggerWaypoint = function (waypointIdentifier, choice) {
        var context = new TriggerWaypointContext();
        context.choice = choice;
        return (this.triggerWaypointInternal(waypointIdentifier, context));
    };
    Engine.prototype.triggerWaypointForce = function (waypointIdentifier, choice) {
        var context = new TriggerWaypointContext();
        context.choice = choice;
        context.force = true;
        return (this.triggerWaypointInternal(waypointIdentifier, context));
    };
    Engine.prototype.triggerWaypointInternal = function (waypointIdentifier, context) {
        context = context.clone();
        waypointIdentifier = waypointIdentifier.split('~')[0];
        if (context.force == null)
            context.force = false;
        if (context.instant == null)
            context.instant = false;
        if (context.seed == null)
            context.seed = this._activeSeed;
        else
            elleLogger.info("context.seed = " + context.seed);
        if (context.time == null)
            context.time = new Date().getTime();
        if (!context.waypoints)
            context.waypoints = this._waypoints;
        if (!context.waypoints) {
            elleLogger.warn("Missing waypoints");
            return false;
        }
        if (!context.playerData)
            context.playerData = this._playerData;
        if (!context.playerData) {
            elleLogger.warn("Missing player data");
            return false;
        }
        if (!context.gameData)
            context.gameData = this._gameData;
        if (!context.gameData) {
            elleLogger.warn("Missing game data");
            return false;
        }
        if (!context.nodeProcessor)
            context.nodeProcessor = this;
        if (!context.nodeProcessor) {
            elleLogger.warn("Missing node processor");
            return false;
        }
        if (!context.categories)
            context.categories = this._categories;
        if (!context.categories) {
            elleLogger.warn("Missing categories");
            return false;
        }
        var tailNode = context.playerData.getLastChoice();
        if (!context.force &&
            tailNode &&
            (tailNode.type !== NodeType.choice) &&
            (tailNode.type !== NodeType.text_input) &&
            (tailNode.type !== NodeType.waypoint)) {
            elleLogger.warn("Not forcing and tail node was not a choice, text_input, or waypoint");
            return false;
        }
        if (!context.force &&
            !tailNode &&
            waypointIdentifier != "Start") {
            elleLogger.warn("First waypoint must be 'Start'");
            return false;
        }
        var waypointNodes = context.waypoints[waypointIdentifier];
        if (!waypointNodes) {
            this._waypointWaitingContext = new WaypointWaitingContext(waypointIdentifier, context);
            requestWaypoint(waypointIdentifier);
            elleLogger.info("No nodes for ''" + waypointIdentifier + "' will make asynchronous request");
            return false;
        }
        if (tailNode && tailNode.type === NodeType.choice) {
            if (context.force || this.isWaypointChoiceValid(waypointIdentifier, tailNode, context.waypoints, context.categories)) {
                tailNode.input_value = waypointIdentifier;
                tailNode.pickedTime = context.time;
            }
            else {
                elleLogger.warn(waypointIdentifier + " was not a valid choice");
                return false;
            }
        }
        if (tailNode && tailNode.type === NodeType.text_input) {
            var textInputNode = tailNode;
            if (typeof (context.choice) !== "string") {
                elleLogger.warn("text_input node at waypoint " + waypointIdentifier + " expected a string for context.choice, not: " + typeof (context.choice));
                return false;
            }
            if (context.force || waypointIdentifier === textInputNode.waypoint) {
                // Sets the story variable
                textInputNode.processInput(context.choice, this, context.gameData);
                // Updates the processedValue of the node
                textInputNode.evaluateNodeWithReturnType(null, this, context.gameData, true);
            }
            else {
                elleLogger.warn(waypointIdentifier + " was not the expected destination of the text_input");
                return false;
            }
        }
        // Update random to seed
        this.updateRandom(context.seed);
        var timeWithOffset = context.time;
        var uiNodes = new Array();
        for (var i = 0; i < waypointNodes.length; ++i) {
            var currentNode = waypointNodes[i];
            if (currentNode.type == NodeType.netcheck) {
                // https://en.wikipedia.org/wiki/Children%27s_Online_Privacy_Protection_Act
                // not a ui node, but needs special processing
                var now = new Date(), year = now.getFullYear(), age = Math.min(Math.max(year - context.gameData.story.age, 0), year);
                context.playerData.age = age;
                if (age >= 13) {
                    // this is so bad....need to change eventually
                    waypointNodes.push(new SequenceNode(NodeType.waypoint, '"enter_email"'));
                }
                else {
                    waypointNodes.push(new SequenceNode(NodeType.waypoint, '"under_age"'));
                }
            }
            var uiNode = context.nodeProcessor.processSequenceNode(currentNode, context.gameData, this._categories);
            if (uiNode) {
                var delay = this.delayForNode(uiNode, context.gameData.gameSpeed, context.playerData.settings.language);
                uiNode.availableTime = timeWithOffset + delay;
                timeWithOffset += delay;
                uiNodes.push(uiNode);
            }
        }
        if (uiNodes.length > 0) {
            elleLogger.info("Adding nodes to player sequence for '" + waypointIdentifier + "':", uiNodes);
        }
        else {
            elleLogger.info("No nodes to add to player sequence for '" + waypointIdentifier + "'");
        }
        context.playerData._sequence = context.playerData._sequence.concat(uiNodes);
        var waypointIndex = context.playerData._simpleSequence.length;
        var simpleWaypoint;
        // Don't add to simple sequence if you are auto-playing a waypoint
        // Start must be in the simple sequence for change language to work properly on an immediate change language situation
        if (tailNode == null ||
            (tailNode && tailNode.type !== NodeType.waypoint) ||
            waypointIdentifier == "Start") {
            simpleWaypoint = new SimpleWaypoint(waypointIdentifier, context.seed, context.time);
            simpleWaypoint.choice = context.choice;
            simpleWaypoint.identifier += "~" + waypointIndex;
            simpleWaypoint.sequenceLength = context.playerData._sequence.length;
            var serializedGameData = context.gameData.serialize();
            simpleWaypoint.serializedGameData = serializedGameData;
            context.playerData._simpleSequence.push(simpleWaypoint);
        }
        else {
            elleLogger.info("No need to add waypoint '" + waypointIdentifier + "' to the simple sequence");
            // We do need to add the sequence nodes in the connected waypoint to the sequence length
            simpleWaypoint = context.playerData._simpleSequence[context.playerData._simpleSequence.length - 1];
            simpleWaypoint.sequenceLength = context.playerData._sequence.length;
            // We also need to serialize the game data into the existing simple sequence node
            var serializedGameData = context.gameData.serialize();
            simpleWaypoint.serializedGameData = serializedGameData;
        }
        var chapters = this.getChapters(false, context);
        if (chapters && chapters.indexOf(waypointIdentifier) >= 0) {
            simpleWaypoint.chapterName = waypointIdentifier;
        }
        var checkpoints = this.getCheckpoints(false, context);
        if (checkpoints && (waypointIdentifier in checkpoints)) {
            simpleWaypoint.chapterName = waypointIdentifier;
        }
        // Grab the last node of the UI sequence and see if it's a waypoint
        var finalNode = context.playerData.getLastChoice();
        if (finalNode && finalNode.isWaypoint()) {
            finalNode.processedValue;
            elleLogger.info("Auto-triggering " + finalNode.processedValue);
            context.choice = undefined;
            context.time = timeWithOffset;
            return this.triggerWaypointInternal(finalNode.processedValue, context);
        }
        else if (finalNode && finalNode.type === NodeType.choice) {
            for (var i = context.playerData._sequence.length - 2; i >= 0; --i) {
                var node = context.playerData._sequence[i];
                if (node.type === NodeType.text) {
                    node.relatedCategory = finalNode.processedValue;
                    break;
                }
            }
        }
        if (finalNode && (finalNode.type === NodeType.choice || finalNode.type === NodeType.text_input)) {
            finalNode.rewindKey = "~" + context.playerData._simpleSequence.length;
        }
        // Update the seed itself
        this.updateRandom();
        return true;
    };
    Engine.prototype.getSequence = function (playerData, fromIndex) {
        if (playerData == null)
            playerData = this._playerData;
        var sequence = playerData.getSequence();
        if (fromIndex && fromIndex < sequence.length) {
            sequence = sequence.slice(fromIndex, sequence.length);
        }
        return SequenceNode.serializeArray(sequence);
    };
    ;
    Engine.prototype.getSequenceSince = function (fromTimeMS) {
        var sequence = this._playerData.getSequence();
        if (fromTimeMS) {
            var cnt = 0;
            for (var idx = sequence.length - 1; idx >= 0; idx--) {
                var node = sequence[idx];
                if (node.availableTime < fromTimeMS) {
                    sequence = sequence.slice(-cnt, sequence.length);
                    break;
                }
                cnt++;
            }
        }
        return SequenceNode.serializeArray(sequence);
    };
    ;
    // This function will process commands in sequence nodes.
    // It is public to support testing.
    Engine.prototype.processSequenceNode = function (sequenceNode, gameData) {
        var resultingNode;
        var processedNode;
        if (!gameData)
            gameData = this._gameData;
        if (!gameData) {
            elleLogger.warn("gameData is null");
            return null;
        }
        if (sequenceNode.type == null) {
            elleLogger.warn("node has no type", sequenceNode);
            return null;
        }
        if (sequenceNode.isControlFlow()) {
            this.processControlFlowNode(sequenceNode, gameData);
            return undefined;
        }
        if (gameData.getIfState() == false) {
            return undefined;
        }
        switch (sequenceNode.type) {
            case NodeType.void:
                sequenceNode.evaluateNodeWithReturnType(null, this, gameData);
                break;
            // String producing non-UI nodes
            case NodeType.dialog:
                processedNode = sequenceNode.evaluateNodeWithReturnType("string", this, gameData);
                gameData.styleKey = processedNode.processedValue;
                break;
            // String producing UI nodes
            case NodeType.text:
            case NodeType.choice:
            case NodeType.game_over:
            case NodeType.image:
            case NodeType.sound:
            case NodeType.music_start:
            case NodeType.music_end:
            case NodeType.text_input:
            case NodeType.waypoint:
            case NodeType.countdown:
            case NodeType.interface:
            case NodeType.endinterface:
            case NodeType.urllink:
            case NodeType.hulkout:
            case NodeType.hulkin:
            case NodeType.placement:
            case NodeType.chapter_end:
            case NodeType.multimedia:
            case NodeType.gamebook_end:
                processedNode = sequenceNode.evaluateNodeWithReturnType("string", this, gameData);
                if (processedNode.processedValue != null) {
                    // Music
                    if (sequenceNode.type === NodeType.music_end) {
                        gameData.music = undefined;
                    }
                    else if (sequenceNode.type === NodeType.music_start) {
                        gameData.music = processedNode.processedValue;
                    }
                    // Dialog
                    // This is not a great pattern, but the styleKey property is
                    // being used to track 'we are a dialog state', so if that
                    // prop exists, handle the wrapping for dialog text
                    if (sequenceNode.type === NodeType.dialog) {
                        elleLogger.info("Dialog: ", processedNode);
                        gameData.styleKey = processedNode.processedValue;
                    }
                    // Add Dialog styling
                    if (sequenceNode.type === NodeType.text) {
                        processedNode.styleKey = gameData.styleKey;
                        if (gameData.styleKey) {
                            processedNode.processedValue = this.wrapDialogNode(processedNode.processedValue);
                        }
                        gameData.styleKey = undefined;
                        elleLogger.info('processedNode', processedNode);
                    }
                    resultingNode = processedNode;
                }
                else if (processedNode.type == 15) {
                    // game over state, no extra processing needed
                    resultingNode = processedNode;
                }
                else {
                    elleLogger.warn("Sequence node did not render string as expected:", JSON.stringify(sequenceNode.serialize()));
                }
                break;
            // Number producing UI nodes
            case NodeType.addelay:
            case NodeType.delay:
                processedNode = sequenceNode.evaluateNodeWithReturnType("number", this, gameData);
                if (processedNode.processedValue != null) {
                    resultingNode = processedNode;
                }
                break;
            case NodeType.netcheck:
                // this is handled elsewhere, discard.
                resultingNode = undefined;
                break;
            case NodeType.sendemail:
                sequenceNode.js = '"' + gameData.story.emailAddress + '";';
                resultingNode = processedNode = sequenceNode.evaluateNodeWithReturnType("string", this, gameData);
                break;
            default:
                elleLogger.warn("Don't know what to do with a node of type '" + sequenceNode.type + "'");
        }
        if (gameData.silently == true && sequenceNode.type != NodeType.sendemail) {
            resultingNode = undefined;
        }
        return resultingNode;
    };
    ;
    Engine.prototype.isWaypointChoiceValid = function (waypointIdentifier, lastChoice, waypoints, categories) {
        if (waypoints[waypointIdentifier] == null) {
            elleLogger.warn("Waypoint identifier " + waypointIdentifier + " is not a known waypoint");
            return false;
        }
        var validChoices = [];
        var activeCategory = categories[lastChoice.processedValue];
        if (activeCategory == null) {
            elleLogger.warn("Category '" + lastChoice.processedValue + "', referenced in previous choice node, does not exist.");
            return false;
        }
        for (var i = 0; i < activeCategory.actions.length; ++i) {
            validChoices.push(activeCategory.actions[i].identifier);
        }
        return (validChoices.indexOf(waypointIdentifier) != -1);
    };
    ;
    Engine.prototype.processControlFlowNode = function (sequenceNode, gameData) {
        var engine = this; // Context for node processing
        if (!gameData)
            gameData = this._gameData;
        switch (sequenceNode.type) {
            case NodeType.if_block:
            case NodeType.else_if_block:
                var ifstate = false;
                try {
                    ifstate = eval(sequenceNode.js);
                }
                catch (e) {
                    elleLogger.warn("Failed to evaluate '" + sequenceNode.js + "': " + e);
                    ifstate = false;
                }
                if (sequenceNode.type === NodeType.if_block) {
                    gameData.startIf(ifstate);
                }
                else {
                    gameData.startElse(ifstate);
                }
                break;
            case NodeType.else_block:
                gameData.startElse(!gameData.getIfState());
                break;
            case NodeType.endif_block:
                gameData.endIf();
                break;
            default:
                elleLogger.warn("This is not a control flow node: " + sequenceNode.type);
        }
    };
    ;
    Engine.prototype.delayForNode = function (node, gameSpeed, language) {
        var delay = 0;
        var charsPerSecond = this.getCharsPerSecond(language);
        switch (node.type) {
            case NodeType.text:
                // Delay is in milliseconds
                delay = 1000 * (Settings.BASE_DELAY_IN_CHARS + node.processedValue.length) / this.getCharsPerSecond(language);
                // Round to nearest millisecond
                delay = Math.round(delay / 10) * 10;
                break;
            case NodeType.choice:
                delay = Settings.CHOICE_NODE_DEFAULT_DELAY; // 1 second
                break;
            case NodeType.delay:
            case NodeType.addelay:
                delay = node.processedValue;
                break;
        }
        switch (gameSpeed) {
            case GameSpeed.Normal:
                // delay = delay; // noop
                break;
            case GameSpeed.Fast:
                delay = Settings.FAST_DELAY;
                break;
            case GameSpeed.SuperFast:
                delay = Settings.SUPER_FAST_DELAY;
                break;
            case GameSpeed.Instant:
                delay = 0;
                break;
        }
        return delay;
    };
    // wrappers to handle different node types
    Engine.prototype.wrapDialogNode = function (value) {
        return value;
    };
    Engine.prototype.getCharsPerSecond = function (language) {
        var charactersPerSecond = (Settings.CHARS_PER_SECOND[language] || Settings.CHARS_PER_SECOND["default"]);
        return charactersPerSecond;
    };
    Engine.prototype.getChapters = function (visitedOnly, context) {
        visitedOnly = (visitedOnly === undefined) ? true : visitedOnly;
        if (context == null)
            context = new TriggerWaypointContext();
        var gameData = (context.gameData == null) ? this._gameData : context.gameData;
        var targets = [];
        var chapters = [];
        if (typeof gameData === 'undefined') {
            return targets;
        }
        if (typeof gameData['chapters'] !== 'undefined') {
            targets = gameData['chapters'];
        }
        else if (typeof gameData['story']['chapters'] !== 'undefined') {
            targets = gameData['story']['chapters'];
        }
        for (var i = 0; i < targets.length; ++i) {
            var target = targets[i];
            if (this.visited(target) || !visitedOnly) {
                chapters.push(target);
            }
        }
        return chapters;
    };
    Engine.prototype.getCheckpoints = function (visitedOnly, context) {
        visitedOnly = (visitedOnly === undefined) ? true : visitedOnly;
        if (context == null)
            context = new TriggerWaypointContext();
        var gameData = (context.gameData == null) ? this._gameData : context.gameData;
        var dictionary = {};
        var checkpoints = {};
        if (typeof gameData === 'undefined') {
            return dictionary;
        }
        if (typeof gameData['checkpoints'] !== 'undefined') {
            dictionary = gameData['checkpoints'];
        }
        else if (typeof gameData['story']['checkpoints'] !== 'undefined') {
            dictionary = gameData['story']['checkpoints'];
        }
        for (var key in dictionary) {
            if (!visitedOnly || this.visited(key)) {
                checkpoints[key] = dictionary[key];
            }
        }
        return checkpoints;
    };
    Engine.prototype.getOrderedCheckpoints = function (context) {
        // returns array of checkpoints in the order they were visited.
        if (context == null)
            context = new TriggerWaypointContext();
        var gameData = (context.gameData == null) ? this._gameData : context.gameData;
        var dictionary = {};
        if (typeof gameData === 'undefined') {
            return [];
        }
        if (typeof gameData['checkpoints'] !== 'undefined') {
            dictionary = gameData['checkpoints'];
        }
        else if (typeof gameData['story']['checkpoints'] !== 'undefined') {
            dictionary = gameData['story']['checkpoints'];
        }
        var waypoints = Object.keys(dictionary); // array of all checkpoint waypoint identifiers
        var playerData = (context.playerData == null) ? this._playerData : context.playerData;
        var sequence = playerData._sequence;
        var checkpoints = [];
        for (var i = 0; i < sequence.length; i++) {
            var sequenceNode = sequence[i];
            var waypoint = null;
            if (sequenceNode.type == NodeType.waypoint && waypoints.indexOf(sequenceNode.processedValue) > -1) {
                waypoint = sequenceNode.processedValue;
            }
            else if (sequenceNode.type == NodeType.choice && waypoints.indexOf(sequenceNode.input_value) > -1) {
                waypoint = sequenceNode.input_value;
            }
            if (waypoint != null) {
                var entry = {};
                entry[waypoint] = dictionary[waypoint];
                checkpoints.push(entry);
                // remove so it's not used again
                var index = waypoints.indexOf(waypoint);
                if (index > -1) {
                    waypoints.splice(index, 1);
                }
            }
            if (waypoints.length == 0) {
                break;
            }
        }
        return checkpoints;
    };
    return Engine;
})(); // Engine
var GameData = (function () {
    function GameData() {
        this._ifstack = [IfElseState.TrueState];
        this.gameSpeed = GameSpeed.Normal;
        this.silently = false;
        this.gameOverSeen = false;
        this.initializeStory();
    }
    GameData.prototype.initializeStory = function () {
        this.story = {};
        this.story.chapters = [];
    };
    GameData.prototype.deserialize = function (input) {
        if (input == null)
            return this;
        if (input.story != null) {
            this.story = input.story;
        }
        if (input.gameSpeed != null) {
            this.gameSpeed = input.gameSpeed;
        }
        if (input.gameOverSeen != null) {
            this.gameOverSeen = input.gameOverSeen;
        }
        return this;
    };
    GameData.prototype.serialize = function () {
        elleLogger.info('GameData.serialize start');
        var output = {};
        output.gameSpeed = this.gameSpeed;
        output.gameOverSeen = this.gameOverSeen;
        // If we ever need to catch function definitions, we could iterate the story
        // looking for functions. We could call "toString()" on the functions and save
        // them specially. We could then eval the definitions as part of deserializing.
        output.story = JSON.parse(JSON.stringify(this.story));
        elleLogger.info('GameData.serialize end');
        return output;
    };
    /**
     * Gets or sets the value of the gameOverSeen flag.
     *
     * Wrapped up in one method since writting massive amounts of getters and
     * setters gets tedious.
     *
     * @param  {boolean} newState Optional. New value of the gameOverSeen flag.
     * @return {boolean}          The current value of gameOverSeen after the optional update
     */
    GameData.prototype.gameOverState = function (newState) {
        if (newState !== undefined) {
            this.gameOverSeen = newState;
        }
        return this.gameOverSeen;
    };
    GameData.prototype.isTrue = function (ifElseState) {
        return ifElseState === IfElseState.TrueState;
    };
    GameData.prototype.getIfState = function () {
        var result = true;
        for (var i = 0; i < this._ifstack.length; ++i) {
            result = result && this.isTrue(this._ifstack[i]);
        }
        return result;
    };
    GameData.prototype.startIf = function (value) {
        if (value) {
            this._ifstack.push(IfElseState.TrueState);
        }
        else {
            this._ifstack.push(IfElseState.NotTrueYet);
        }
    };
    GameData.prototype.startElse = function (value) {
        var ifElseState = this._ifstack[this._ifstack.length - 1];
        switch (ifElseState) {
            case IfElseState.NotTrueYet:
                if (value) {
                    this._ifstack.pop();
                    this._ifstack.push(IfElseState.TrueState);
                }
                break;
            case IfElseState.TrueState:
                this._ifstack.pop();
                this._ifstack.push(IfElseState.FalseState);
                break;
            case IfElseState.FalseState:
                break;
        }
    };
    GameData.prototype.endIf = function () {
        this._ifstack.pop();
    };
    return GameData;
})();
var GameSpeed;
(function (GameSpeed) {
    GameSpeed[GameSpeed["Normal"] = 0] = "Normal";
    GameSpeed[GameSpeed["Fast"] = 1] = "Fast";
    GameSpeed[GameSpeed["SuperFast"] = 2] = "SuperFast";
    GameSpeed[GameSpeed["Instant"] = 3] = "Instant";
})(GameSpeed || (GameSpeed = {}));
;
var elleLogger = (this.native_logger || { info: function () { }, warn: function () { }, error: function () { } });
var createEngine = function () {
    return ((engine = new Engine()) != null);
};
// A stub waypoint requester for the web
var ecma_requestWaypoint = function (waypointIdentifier) {
    elleLogger.warn("Requested unknown waypoint: " + waypointIdentifier);
};
// Use the native waypoint requester, if available.
var requestWaypoint = (this.native_requestWaypoint || ecma_requestWaypoint);
var IfElseState;
(function (IfElseState) {
    IfElseState[IfElseState["NotTrueYet"] = 0] = "NotTrueYet";
    IfElseState[IfElseState["TrueState"] = 1] = "TrueState";
    IfElseState[IfElseState["FalseState"] = 2] = "FalseState";
})(IfElseState || (IfElseState = {}));
;
var NodeType;
(function (NodeType) {
    NodeType[NodeType["text"] = 0] = "text";
    NodeType[NodeType["void"] = 1] = "void";
    NodeType[NodeType["choice"] = 2] = "choice";
    NodeType[NodeType["delay"] = 3] = "delay";
    NodeType[NodeType["image"] = 4] = "image";
    NodeType[NodeType["if_block"] = 5] = "if_block";
    NodeType[NodeType["else_if_block"] = 6] = "else_if_block";
    NodeType[NodeType["else_block"] = 7] = "else_block";
    NodeType[NodeType["endif_block"] = 8] = "endif_block";
    NodeType[NodeType["text_input"] = 9] = "text_input";
    NodeType[NodeType["sound"] = 10] = "sound";
    NodeType[NodeType["music_start"] = 11] = "music_start";
    NodeType[NodeType["music_end"] = 12] = "music_end";
    NodeType[NodeType["waypoint"] = 13] = "waypoint";
    NodeType[NodeType["dialog"] = 14] = "dialog";
    NodeType[NodeType["game_over"] = 15] = "game_over";
    NodeType[NodeType["netcheck"] = 16] = "netcheck";
    NodeType[NodeType["sendemail"] = 17] = "sendemail";
    NodeType[NodeType["countdown"] = 18] = "countdown";
    NodeType[NodeType["interface"] = 19] = "interface";
    NodeType[NodeType["endinterface"] = 20] = "endinterface";
    NodeType[NodeType["urllink"] = 21] = "urllink";
    NodeType[NodeType["hulkout"] = 22] = "hulkout";
    NodeType[NodeType["hulkin"] = 23] = "hulkin";
    NodeType[NodeType["placement"] = 24] = "placement";
    NodeType[NodeType["chapter_end"] = 25] = "chapter_end";
    NodeType[NodeType["multimedia"] = 26] = "multimedia";
    NodeType[NodeType["addelay"] = 27] = "addelay";
    NodeType[NodeType["gamebook_end"] = 28] = "gamebook_end";
})(NodeType || (NodeType = {}));
;
var PlayerData = (function () {
    function PlayerData() {
        this.playerInput = {};
        this.purchased = {}; //purchase list
        this._sequence = [];
        this._simpleSequence = [];
        this.settings = new Settings();
    }
    PlayerData.prototype.setLanguage = function (langauge) {
        this.settings.setLanguage(langauge);
        return;
    };
    PlayerData.prototype.deserialize = function (input) {
        elleLogger.info('PlayerData.deserialize start');
        if (input == null)
            return this;
        if (input.sequence != null) {
            var sequence = [];
            for (var i = 0; i < input.sequence.length; ++i) {
                var sequenceNode = SequenceNode.nodeFromSerializedObject(input.sequence[i]);
                if (sequenceNode) {
                    sequence.push(sequenceNode);
                }
            }
            this._sequence = sequence;
        }
        if (input.simpleSequence != null)
            this._simpleSequence = input.simpleSequence;
        this.settings = new Settings().deserialize(input.settings);
        if (input.playerInput != null)
            this.playerInput = input.playerInput;
        if (input.purchased != null)
            this.purchased = input.purchased;
        if (input.age != null)
            this.age = input.age;
        if (input.email != null)
            this.email = input.email;
        elleLogger.info('PlayerData.deserialize end');
        return this;
    };
    PlayerData.prototype.serialize = function () {
        elleLogger.info('PlayerData.serialize start');
        var output = {};
        output.sequence = [];
        elleLogger.info('PlayerData.serialize sequencing start', 'length ' + this._sequence.length);
        for (var i = 0; i < this._sequence.length; ++i) {
            output.sequence.push(this._sequence[i].serialize());
        }
        elleLogger.info('PlayerData.serialize sequencing end');
        output.simpleSequence = this._simpleSequence;
        output.settings = this.settings;
        output.playerInput = this.playerInput;
        output.purchased = this.purchased;
        output.age = this.age;
        output.email = this.email;
        elleLogger.info('PlayerData.serialize start');
        return output;
    };
    PlayerData.prototype.getLastChoice = function () {
        var length = this._sequence.length;
        if (length > 0) {
            return this._sequence[length - 1];
        }
        return null;
    };
    PlayerData.prototype.getSequence = function () {
        return this._sequence;
    };
    return PlayerData;
})();
;
var RewindContext = (function () {
    function RewindContext() {
    }
    RewindContext.prototype.clone = function () {
        var clone = new RewindContext();
        clone.categories = this.categories;
        clone.gameData = this.gameData;
        clone.playerData = this.playerData;
        clone.replay = this.replay;
        clone.waypoints = this.waypoints;
        return clone;
    };
    return RewindContext;
})();
var SequenceNode = (function () {
    function SequenceNode(aType, aJs) {
        this.type = aType;
        this.js = aJs;
    }
    /**
    * Needed due to duktape issue: https://github.com/svaarala/duktape/issues/427
    *
    * @param  {number} index The index of the enum value
    * @return {string}       The name of the enum at the given index
    */
    SequenceNode.getNodeTypeName = function (index) {
        var name;
        switch (index) {
            case NodeType.text:
                name = "text";
                break;
            case NodeType.void:
                name = "void";
                break;
            case NodeType.choice:
                name = "choice";
                break;
            case NodeType.delay:
                name = "delay";
                break;
            case NodeType.dialog:
                name = "dialog";
                break;
            case NodeType.game_over:
                name = "game_over";
                break;
            case NodeType.image:
                name = "image";
                break;
            case NodeType.if_block:
                name = "if_block";
                break;
            case NodeType.else_if_block:
                name = "else_if_block";
                break;
            case NodeType.else_block:
                name = "else_block";
                break;
            case NodeType.endif_block:
                name = "endif_block";
                break;
            case NodeType.text_input:
                name = "text_input";
                break;
            case NodeType.sound:
                name = "sound";
                break;
            case NodeType.music_start:
                name = "music_start";
                break;
            case NodeType.music_end:
                name = "music_end";
                break;
            case NodeType.waypoint:
                name = "waypoint";
                break;
            case NodeType.netcheck:
                name = "netcheck";
                break;
            case NodeType.sendemail:
                name = "sendemail";
                break;
            case NodeType.countdown:
                name = "countdown";
                break;
            case NodeType.interface:
                name = "interface";
                break;
            case NodeType.endinterface:
                name = "endinterface";
                break;
            case NodeType.urllink:
                name = "urllink";
                break;
            case NodeType.hulkout:
                name = "hulkout";
                break;
            case NodeType.hulkin:
                name = "hulkin";
                break;
            case NodeType.placement:
                name = "placement";
                break;
            case NodeType.chapter_end:
                name = "chapter_end";
                break;
            case NodeType.multimedia:
                name = "multimedia";
                break;
            case NodeType.addelay:
                name = "addelay";
                break;
            case NodeType.gamebook_end:
                name = "gamebook_end";
                break;
        }
        return name;
    };
    SequenceNode.nodeFromSerializedObject = function (input) {
        var node;
        var type = NodeType[input.type];
        if (type === NodeType.text_input) {
            node = new TextInputNode().deserialize(input);
        }
        else {
            node = new SequenceNode().deserialize(input);
        }
        return node;
    };
    // Being able to pass an object to clone() simplifies inheritance
    SequenceNode.prototype.clone = function (clone) {
        if (clone == null)
            clone = new SequenceNode();
        clone.type = this.type;
        clone.js = this.js;
        clone.availableTime = this.availableTime;
        clone.pickedTime = this.pickedTime;
        clone.input_value = this.input_value;
        clone.rewindKey = this.rewindKey;
        clone.processedValue = this.processedValue;
        clone.relatedCategory = this.relatedCategory;
        clone.styleKey = this.styleKey;
        clone.scratchpad = this.scratchpad;
        return clone;
    };
    SequenceNode.prototype.deserialize = function (input) {
        var type = NodeType[input.type];
        this.type = type;
        this.js = input.js;
        this.availableTime = input.availableTime;
        this.pickedTime = input.pickedTime;
        this.input_value = input.input_value;
        this.rewindKey = input.rewindKey;
        this.processedValue = input.processedValue;
        this.relatedCategory = input.relatedCategory;
        this.styleKey = input.styleKey;
        this.scratchpad = input.scratchpad;
        return this;
    };
    SequenceNode.serializeArray = function (sequence) {
        var serializedSequence = new Array();
        for (var i = 0; i < sequence.length; ++i) {
            serializedSequence.push(sequence[i].serialize());
        }
        return serializedSequence;
    };
    SequenceNode.prototype.serialize = function () {
        var output = {};
        // Temporary fix for duktape
        output.type = SequenceNode.getNodeTypeName(this.type);
        // output.type = NodeType[this.type];
        output.js = this.js;
        output.availableTime = this.availableTime;
        output.pickedTime = this.pickedTime;
        output.input_value = this.input_value;
        output.rewindKey = this.rewindKey;
        output.processedValue = this.processedValue;
        output.relatedCategory = this.relatedCategory;
        output.styleKey = this.styleKey;
        output.scratchpad = this.scratchpad;
        return output;
    };
    SequenceNode.prototype.isWaypoint = function () {
        return this.type === NodeType.waypoint;
    };
    SequenceNode.prototype.isControlFlow = function () {
        var node = this;
        var isControlFlow = false;
        isControlFlow = isControlFlow || (node.type === NodeType.if_block);
        isControlFlow = isControlFlow || (node.type === NodeType.else_if_block);
        isControlFlow = isControlFlow || (node.type === NodeType.else_block);
        isControlFlow = isControlFlow || (node.type === NodeType.endif_block);
        return isControlFlow;
    };
    /**
     * This method does the magic. It runs the JS provided in the node.
     *
     * @param  {SequenceNode} sequenceNode Node to process.
     * @param  {string}       returnType   Type name, can be null. (e.g. "string")
     * @param  {GameData}     gameData     This object is used for context.
     * @return {SequenceNode}              Returns a new node with the processedValue set, if successful
     */
    SequenceNode.prototype.evaluateNodeWithReturnType = function (returnType, engine, gameData, inPlace) {
        var node = inPlace ? this : this.clone();
        var processedValue;
        try {
            processedValue = this.js ? eval(this.js) : '';
            if (returnType != null && typeof (processedValue) !== returnType) {
                elleLogger.warn("Evaluation of JS '" + this.js + "' did not produce a " + returnType + ".");
                processedValue = null;
            }
        }
        catch (e) {
            elleLogger.warn("Failed to evaluate '" + this.js + "': " + e);
        }
        node.processedValue = processedValue;
        return node;
    };
    return SequenceNode;
})();
var Settings = (function () {
    function Settings(aGameSpeed, aLanguage) {
        if (aGameSpeed == null) {
            this.gameSpeed = GameSpeed.Normal;
        }
        else {
            this.gameSpeed = aGameSpeed;
        }
        if (aLanguage == null) {
            this.language = "en";
        }
        else {
            this.language = aLanguage;
        }
    }
    /**
    * Needed due to duktape issue: https://github.com/svaarala/duktape/issues/427
    *
    * @param  {number} index The index of the enum value
    * @return {string}       The name of the enum at the given index
    */
    Settings.getGameSpeedName = function (index) {
        var name;
        switch (index) {
            case GameSpeed.Normal:
                name = "Normal";
                break;
            case GameSpeed.Fast:
                name = "Fast";
                break;
            case GameSpeed.SuperFast:
                name = "SuperFast";
                break;
            case GameSpeed.Instant:
                name = "Instant";
                break;
        }
        return name;
    };
    Settings.prototype.setLanguage = function (language) {
        this.language = language;
        return;
    };
    Settings.prototype.deserialize = function (input) {
        if (input == null)
            return this;
        if (input.gameSpeed != null)
            this.gameSpeed = GameSpeed[input.gameSpeed];
        if (input.language != null)
            this.language = input.language;
        return this;
    };
    Settings.prototype.serialize = function () {
        var output;
        output.gameSpeed = GameSpeed[this.gameSpeed];
        output.language = this.language;
        return output;
    };
    Settings.FAST_DELAY = 1000;
    Settings.SUPER_FAST_DELAY = 100;
    Settings.CHOICE_NODE_DEFAULT_DELAY = 1000;
    Settings.CHARS_PER_SECOND = { "ja": 10, "zh": 10, "default": 30 };
    Settings.BASE_DELAY_IN_CHARS = 30;
    return Settings;
})();
var SimpleWaypoint = (function () {
    function SimpleWaypoint(aIdentifier, aSeed, aActivationEpoch) {
        this.identifier = aIdentifier;
        this.seed = aSeed;
        this.activationEpoch = aActivationEpoch;
    }
    return SimpleWaypoint;
})();
var __extends = (this && this.__extends) || function (d, b) {
    for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p];
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
};
var TextInputNode = (function (_super) {
    __extends(TextInputNode, _super);
    function TextInputNode(aType, aJs) {
        _super.call(this, NodeType.text_input, aJs);
    }
    TextInputNode.prototype.clone = function (clone) {
        if (clone == null)
            clone = _super.prototype.clone.call(this, new TextInputNode());
        clone.js_set = this.js_set;
        clone.js_unset = this.js_unset;
        clone.target = this.target;
        clone.waypoint = this.waypoint;
        return clone;
    };
    TextInputNode.prototype.deserialize = function (input) {
        _super.prototype.deserialize.call(this, input);
        this.js_set = input.js_set;
        this.js_unset = input.js_unset;
        this.target = input.target;
        this.waypoint = input.waypoint;
        return this;
    };
    TextInputNode.prototype.serialize = function () {
        var output = _super.prototype.serialize.call(this);
        output.js_set = this.js_set;
        output.js_unset = this.js_unset;
        output.target = this.target;
        output.waypoint = this.waypoint;
        return output;
    };
    TextInputNode.prototype.processInput = function (input, engine, gameData) {
        input = input.replace(/\"/g, ''); // Sorry, no double quotes allowed! Removed entirely.
        var evalString = this.target + "=\"" + input + "\";";
        try {
            eval(evalString);
            this.input_value = input;
        }
        catch (e) {
            elleLogger.warn("Failed to eval " + evalString + ": " + e);
        }
    };
    TextInputNode.prototype.evaluateNodeWithReturnType = function (returnType, engine, gameData, inPlace) {
        var node = inPlace ? this : this.clone();
        node.js = (node.input_value == null) ? node.js_unset : node.js_set;
        SequenceNode.prototype.evaluateNodeWithReturnType.call(node, "string", engine, gameData, true);
        return node;
    };
    return TextInputNode;
})(SequenceNode);
var TriggerWaypointContext = (function () {
    function TriggerWaypointContext() {
    }
    TriggerWaypointContext.prototype.clone = function () {
        var clone = new TriggerWaypointContext();
        clone.choice = this.choice;
        clone.categories = this.categories;
        clone.force = this.force;
        clone.gameData = this.gameData;
        clone.instant = this.instant;
        clone.nodeProcessor = this.nodeProcessor;
        clone.playerData = this.playerData;
        clone.seed = this.seed;
        clone.time = this.time;
        clone.waypoints = this.waypoints;
        return clone;
    };
    return TriggerWaypointContext;
})();
var WaypointWaitingContext = (function () {
    function WaypointWaitingContext(waypointIdentifier, context) {
        this.context = context;
        this.waypointIdentifier = waypointIdentifier;
    }
    return WaypointWaitingContext;
})();
//# sourceMappingURL=engine.js.map